/*

 Le sPin - A multiplayer game for saturn
 Copyright (C) 2003 SSMTF

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PATICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */
   /*
     objects : 
     station
     tail
     particle
     player 
     camera 
     deco
     arena
     speed_marks
     life_bar
     light
   */ 

#include    "sl_def.h"
#include    "sgl.h"
#include    "scrltxt.h"

#include "mod_att.h"
//#include "dog.h"

///////////////////////////////////////////SOUND

#include    "sddrvs.dat"

#define     SoundSeqBuf		0x25a0b000
extern char sound_map[];
extern char sound_dat[];
extern Uint32 mapsize;
extern Uint32 datsize;

void init_sound(){
  slInitSound(sddrvstsk , sizeof(sddrvstsk) , (Uint8 *)sound_map , mapsize) ;
  slDMACopy(sound_dat , (void *)(SoundSeqBuf) , datsize) ;
}



typedef enum _snd_type {snd_lespin,snd_birth ,snd_death , snd_good ,snd_bad,snd_bing/*,snd_accel*/} snd_type;

//void sound_play(Uint16 num,Uint8 prio, Uint8 vol, Uint8 Pan){
void sound_play(snd_type sfx){ 

  switch (sfx){
  case snd_lespin : 
    slSequenceOn((0 << 8) + 0,31,127,0);
    //    scrollTxtAddTxtLine("snd_lespin");
    break;
  case snd_birth : 
    slSequenceOn((0 << 8) + 1,31,127,0);
    //    scrollTxtAddTxtLine("snd_birth");
   break;
  case snd_death : 
    slSequenceOn((0 << 8) + 2,31,127,0);
    //    scrollTxtAddTxtLine("snd_death");
    break;
  case snd_good : 
    slSequenceOn((0 << 8) + 3,15,127,0);
    //    scrollTxtAddTxtLine("snd_good");
    break;
  case snd_bad : 
    slSequenceOn((0 << 8) + 4,15,127,0);
    //    scrollTxtAddTxtLine("snd_bad");
    break;
  case snd_bing : 
    slSequenceOn((0 << 8) + 5,1,127,0);
    //scrollTxtAddTxtLine("snd_bing");
    break;
    /*  case snd_accel : 
	slSequenceOn((0 << 8) + 6,0,127,0);
	// scrollTxtAddTxtLine("snd_accel");
	break;
    */
  default:
    break;
  }

}

/////////////////////////////////////////////////

#define FALSE (1!=1)
#define TRUE  (1==1)

#define N_PARTICLE 9
#define N_PLAYER   6
#define N_STATION  (N_PLAYER+1)

#define S_SIZE 100.0 //2000
#define INFINITE (toFIXED(S_SIZE*10.0))
#define S_Xm toFIXED(-S_SIZE)
#define S_Ym toFIXED(-S_SIZE)
#define S_Zm toFIXED(-S_SIZE)
#define S_XM toFIXED(S_SIZE)
#define S_YM toFIXED(S_SIZE)
#define S_ZM toFIXED(S_SIZE)

#define MAX_SPEED toFIXED(S_SIZE/10.0)
#define MAX_PLAYER_ENERGY toFIXED(15.0)

#define CO_1_p C_RGB(31,0,0) // primary color for particle 1
#define CO_1_s C_RGB(31,0,5) // secondary ..................

#define CO_2_p C_RGB(0,31,0)
#define CO_2_s C_RGB(0,31,5)

#define CO_3_p C_RGB(0,0,31)
#define CO_3_s C_RGB(5,0,31)                                                                                   
#define CO_4_p C_RGB(31,31,0)
#define CO_4_s C_RGB(31,31,5)

#define CO_5_p C_RGB(31,0,31)
#define CO_5_s C_RGB(31,5,31)                      

#define CO_6_p C_RGB(0,31,31)
#define CO_6_s C_RGB(5,31,31)

#define CO_n_p C_RGB(20,20,30) // color for everybody
#define CO_n_s C_RGB(25,20,30)


/* 
 * RANDOM 
 */

FIXED minmaxrnd(FIXED min, FIXED max){
  FIXED resu,mmax,mmin;
  mmax = (max>min)?max:min;
  mmin = (max>min)?min:max;
  resu = min + slMulFX(slRandom(),max-min);
  return resu;
}

/*
 * PAD
 */

// 1->0
Uint16 pad_GetDownFront (Uint16 prev, Uint16 cur) {
  Uint16 downfront;
  downfront = (~((~prev) & cur));  
  return downfront;
}

// 0->1
Uint16 pad_GetUpFront (Uint16 prev, Uint16 cur) {
  Uint16 upfront = 0;
  upfront = pad_GetDownFront(cur,prev);
  return upfront;
}

/*
 * SPEED MARKS
 */
static POINT point_SPEED_MARKS[]={
  POStoFIXED(-S_SIZE-10.0, -S_SIZE-10.0,  S_SIZE/10.0),
  POStoFIXED( S_SIZE+10.0, -S_SIZE-10.0,  S_SIZE/10.0),
  POStoFIXED( S_SIZE+10.0,  S_SIZE+10.0,  S_SIZE/10.0),
  POStoFIXED(-S_SIZE-10.0,  S_SIZE+10.0,  S_SIZE/10.0),
  POStoFIXED(-S_SIZE-10.0, -S_SIZE-10.0, -S_SIZE/10.0),
  POStoFIXED( S_SIZE+10.0, -S_SIZE-10.0, -S_SIZE/10.0),
  POStoFIXED( S_SIZE+10.0,  S_SIZE+10.0, -S_SIZE/10.0),
  POStoFIXED(-S_SIZE-10.0,  S_SIZE+10.0, -S_SIZE/10.0),
};

POLYGON polygon_SPEED_MARKS[]={
  NORMAL(-1.000000, 0.000000, 0.000000), VERTICES(4,0,3,7),
  NORMAL( 1.000000, 0.000000, 0.000000), VERTICES(1,5,6,2),
};
/*
  ATTR  attribute_SPEED_MARKS[] = { 
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(25,12,25),GRtbl(0),CL32KRGB|MESHoff|CL_Gouraud,sprPolygon, UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(12,31,12),GRtbl(3),CL32KRGB|MESHoff|CL_Gouraud,sprPolygon, UseLight),
  };
*/

ATTR  attribute_SPEED_MARKS[] = { /* */
 ATTRIBUTE(Dual_Plane, SORT_CEN, No_Texture,C_RGB(25,12,25), No_Gouraud, CL32KRGB|MESHoff, sprPolyLine , UseLight),
 ATTRIBUTE(Dual_Plane, SORT_CEN, No_Texture,C_RGB(12,31,12), No_Gouraud, CL32KRGB|MESHoff, sprPolyLine , UseLight),
};

XPDATA	pdata_SPEED_MARKS  = {
  point_SPEED_MARKS, sizeof(point_SPEED_MARKS)/sizeof(POINT),
  polygon_SPEED_MARKS, sizeof(polygon_SPEED_MARKS)/sizeof(POLYGON),
  attribute_SPEED_MARKS,
  NULL,
};

typedef struct _SPEED_MARKS {
  XPDATA* gfx;
  FIXED  sizeZ;       // size in Z of the graphical element
  FIXED first_pos;
  FIXED step_size;
  FIXED maxZ,minZ;
} SPEED_MARKS;

void speed_marks_init(SPEED_MARKS* sm, FIXED sizeZ, FIXED step_size){
  sm->gfx = &pdata_SPEED_MARKS ;
  // sm->sizeZ = toFIXED((2.0*S_SIZE)/10.0);
  sm->sizeZ = sizeZ; //toFIXED((2.0*S_SIZE)/10.0);
  sm->first_pos = toFIXED(0.0);
  //  sm->step_size = toFIXED(S_SIZE/10.0/10.0);
  sm->step_size = step_size;// toFIXED(S_SIZE/15.0);
  sm->maxZ = toFIXED(S_SIZE*5.0);
  sm->minZ = toFIXED(S_SIZE*-5.0);
}

void speed_marks_set_step_size(SPEED_MARKS* sm, FIXED step_size){
  sm->step_size = step_size;
}

void speed_marks_set_sizeZ(SPEED_MARKS* sm, FIXED sizeZ){
  sm->sizeZ = sizeZ;
}

void speed_marks_step(SPEED_MARKS* sm){
  sm->first_pos -= sm->step_size;
  if (sm->first_pos <= toFIXED(0.0)) sm->first_pos += slMulFX(sm->sizeZ,toFIXED(4.0));
}

void speed_marks_display(SPEED_MARKS* sm){
  FIXED pos;
  int rot = TRUE;

  pos = sm->minZ + sm->first_pos;
  //pos += slMulFX(toFIXED(S_SIZE/20.0),slRandom());
  while (pos <= sm->maxZ){
    rot = !rot;
    slPushMatrix();
    {
      slTranslate(toFIXED(0.0), toFIXED(0.0),pos);
      slPushMatrix();
      {
	FIXED l[] = {toFIXED(0.5),toFIXED(0.2),toFIXED(0.3)};
	slRotZ(rot?DEGtoANG(0.0):DEGtoANG(90.0));
	slPutPolygonX(sm->gfx,l);
      }
      slPopMatrix();
    }
    slPopMatrix();

    pos += sm->sizeZ;
  }
}

/*
 * WALL DECO
 */

POINT point_DECO[] = {
  POStoFIXED( 0.0000, -1.0000, -1.0000),
  POStoFIXED( 1.0000, -1.0000, -1.0000),
  POStoFIXED( 1.0000,  1.0000, -1.0000),
  POStoFIXED( 0.0000,  1.0000, -1.0000),
  POStoFIXED( 0.0000, -1.0000,  1.0000),
  POStoFIXED( 1.0000, -1.0000,  1.0000),
  POStoFIXED( 1.0000,  1.0000,  1.0000),
  POStoFIXED( 0.0000,  1.0000,  1.0000),
};

POLYGON	polygon_DECO[] = {
  NORMAL(  0.0000,  0.0000, -1.0000), VERTICES( 0, 1, 2, 3),
  NORMAL(  1.0000,  0.0000,  0.0000), VERTICES( 1, 5, 6, 2),
  NORMAL(  0.0000,  0.0000,  1.0000), VERTICES( 5, 4, 7, 6),
  NORMAL( -1.0000,  0.0000,  0.0000), VERTICES( 4, 0, 3, 7),
  NORMAL(  0.0000, -1.0000,  0.0000), VERTICES( 0, 4, 5, 1),
  NORMAL(  0.0000,  1.0000,  0.0000), VERTICES( 2, 6, 7, 3),
};

ATTR	attribute_DECO[] = { /*sprPolygon */

  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_1_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine , UseLight),
  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_2_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_3_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_4_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_5_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Single_Plane, SORT_CEN, No_Texture, CO_6_p, No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
};

PDATA	pdata_DECO  = {
  point_DECO, sizeof(point_DECO)/sizeof(POINT),
  polygon_DECO, sizeof(polygon_DECO)/sizeof(POLYGON),
  attribute_DECO,
};

void deco_display_a_deco(){
  slPushMatrix();
  {
    slTranslate(S_XM, toFIXED(0.0),toFIXED(0.0));
    slPushMatrix();
    {
      slScale(toFIXED(1.0),S_YM,S_ZM);
      //      slPutPolygon(&pdata_WALL);
    }
    slPopMatrix();
  }
  slPopMatrix();
}

void deco_display(/*DECO* de*/){
  // foreach deco...
  slPushMatrix(); // fond
  {
    slRotY(DEGtoANG(0.0));
    deco_display_a_deco(); 
  }
  slPopMatrix();
  slPushMatrix(); // fond
  {
    slRotY(DEGtoANG(180.0));
    deco_display_a_deco();
  }
  slPopMatrix();
  slPushMatrix(); // fond
  {
    slRotY(DEGtoANG(90.0));
    deco_display_a_deco();
  }
  slPopMatrix();
  slPushMatrix(); // fond
  {
    slRotY(DEGtoANG(270.0));
    deco_display_a_deco();
  }
  slPopMatrix();
  slPushMatrix(); // fond
  {
    slRotZ(DEGtoANG(90.0));
    deco_display_a_deco();
  }
  slPopMatrix();
  slPushMatrix(); // fond
  {
    slRotZ(DEGtoANG(-90.0));
    deco_display_a_deco();
  }
  slPopMatrix();
}

/*
 * ARENA
 */
static POINT point_ARENA[]={
  POStoFIXED(-S_SIZE, -S_SIZE,  S_SIZE),
  POStoFIXED( S_SIZE, -S_SIZE,  S_SIZE),
  POStoFIXED( S_SIZE,	S_SIZE,  S_SIZE),
  POStoFIXED(-S_SIZE,	S_SIZE,  S_SIZE),
  POStoFIXED(-S_SIZE, -S_SIZE, -S_SIZE),
  POStoFIXED( S_SIZE, -S_SIZE, -S_SIZE),
  POStoFIXED( S_SIZE,	S_SIZE, -S_SIZE),
  POStoFIXED(-S_SIZE,	S_SIZE, -S_SIZE),
};

POLYGON polygon_ARENA[]={
  NORMAL( 0.000000, 0.000000, 1.000000), VERTICES(0,1,2,3),
  NORMAL(-1.000000, 0.000000, 0.000000), VERTICES(4,0,3,7),
  NORMAL( 0.000000, 0.000000,-1.000000), VERTICES(5,4,7,6),
  NORMAL( 1.000000, 0.000000, 0.000000), VERTICES(1,5,6,2),
  NORMAL( 0.000000,-1.000000, 0.000000), VERTICES(4,5,1,0),
  NORMAL( 0.000000, 1.000000, 0.000000), VERTICES(3,2,6,7),
};


ATTR	attribute_ARENA[] = { /*sprPolygon */

  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,31,31),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine , UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31, 0, 0),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB( 0,31, 0),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB( 0, 0,31),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,31, 0),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB( 0,31,31),No_Gouraud, CL32KRGB|MESHoff, sprPolyLine, UseLight),

  /*
    ATTRIBUTE(Single_Plane, SORT_CEN, 0,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
    ATTRIBUTE(Single_Plane, SORT_CEN, 1,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
    ATTRIBUTE(Single_Plane, SORT_CEN, 0,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
    ATTRIBUTE(Single_Plane, SORT_CEN, 1,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
    ATTRIBUTE(Single_Plane, SORT_CEN, 0,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
    ATTRIBUTE(Single_Plane, SORT_CEN, 1,No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),
  */
  /*  ATTRIBUTE(Single_Plane, SORT_CEN, PN_SONIC, No_Palet, No_Gouraud, CL32KRGB|MESHoff, sprNoflip, No_Option),*/

};

PDATA	pdata_ARENA  = {
  point_ARENA, sizeof(point_ARENA)/sizeof(POINT),
  polygon_ARENA, sizeof(polygon_ARENA)/sizeof(POLYGON),
  attribute_ARENA,
};


void arena_display(/*AREA* ar*/){
  slPutPolygon(&pdata_ARENA);
  /*
    slPushMatrix();
    {
    slScale(toFIXED(1.02),toFIXED(1.02),toFIXED(1.02));
    slPushMatrix();
    {
    slRotX(180.0);
    slPutPolygon(&pdata_ARENA);
    }
    slPopMatrix();
    }
    slPopMatrix();
  */
}

/*
 * STATIONS
 */
POINT point_STATION[]= {
  POStoFIXED(-0.5,-0.5,0.0),
  POStoFIXED( 0.5,-0.5,0.0),
  POStoFIXED( 0.5, 0.5,0.0),
  POStoFIXED(-0.5, 0.5,0.0),
};

POLYGON polygon_STATION[]= {
  NORMAL(0.0,0.0,1.0),
  VERTICES(0,1,2,3),
};

ATTR attribute_STATION1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION1_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION2_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION3_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION4_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION5_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION6[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION6_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};
ATTR attribute_STATION7[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_n_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};
ATTR attribute_STATION7_line[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_n_p,No_Gouraud,MESHoff,sprPolyLine,UseLight),  
};

PDATA PD_STATION1 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION1
};
PDATA PD_STATION1_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION1_line
};
PDATA PD_STATION2 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION2
};
PDATA PD_STATION2_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION2_line
};
PDATA PD_STATION3 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION3
};
PDATA PD_STATION3_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION3_line
};
PDATA PD_STATION4 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION4
};
PDATA PD_STATION4_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION4_line
};
PDATA PD_STATION5 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION5
};
PDATA PD_STATION5_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION5_line
};
PDATA PD_STATION6 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION6
};
PDATA PD_STATION6_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION6_line
};
PDATA PD_STATION7 = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION7
};
PDATA PD_STATION7_line = {
  point_STATION,sizeof(point_STATION)/sizeof(POINT),
  polygon_STATION,sizeof(polygon_STATION)/sizeof(POLYGON),
  attribute_STATION7_line
};
typedef enum _station_type {for_all, for_one} station_type;
typedef enum _station_state {away, coming, here, leaving} station_state;

typedef struct _STATION {
  station_type type;
  int num;               // corresponding player num
  //  FIXED lifetime;        // in synch() number
  station_state state;   
  FIXED r_time;           // in synch() number. For all state machine purposes
  FIXED pos[XYZ];
  FIXED zspeed;
  FIXED size[XYZ];
  ANGLE ang[XYZ];
  PDATA* gfx;
  PDATA* gfx_trans; // tranparent 
} STATION ;

void station_init(STATION* st, station_type type, int num, PDATA* gfx,PDATA* gfx_trans, FIXED zspeed){
  st->type = type;
  if (type == for_one){
    st->num = num;
  } 
  st->ang[X] = st->ang[Y] = st->ang[Z] = DEGtoANG(0.0);
  st->size[X] = st->size[Y] = st->size[Z] = toFIXED(1.0);
  st->zspeed = zspeed;
  st->gfx = gfx;
  st->gfx_trans  = gfx_trans;
  st->state = away;
  st->r_time = minmaxrnd(toFIXED(50.0),toFIXED(3000.0));
}

void station_set_pseudo_speed(STATION* st,FIXED dz){
  st->zspeed = dz;
}

void station_randomize(STATION* st){

  FIXED ms = toFIXED(S_SIZE/20.0); // min size 
  FIXED rnd, tmp;
  FIXED pos1[3];
  FIXED pos2[3];

  // two random points.
  pos1[X] = minmaxrnd(S_Xm+ms, S_XM-ms);
  pos2[X] = minmaxrnd(S_Xm+ms, S_XM-ms);
  pos1[Y] = minmaxrnd(S_Ym+ms, S_YM-ms);
  pos2[Y] = minmaxrnd(S_Ym+ms, S_YM-ms);

  rnd = slRandom(); // four possible localisation types.

  if (rnd > toFIXED(0.75)) { // top
    pos1[Y] = S_Ym;
    //    scrollTxtAddTxtLine("1");
  } else if (rnd > toFIXED(0.5)){ // left
    pos1[X] = S_Xm;
    //    scrollTxtAddTxtLine("2");
  } else if (rnd > toFIXED(0.25)){ // bottom
    pos1[Y] = S_YM;
    //    scrollTxtAddTxtLine("3");
  } else { // right
    pos1[X] = S_XM;
    //    scrollTxtAddTxtLine("4");
  }

  // reorder
  if (pos1[X]>pos2[X]){tmp = pos2[X];pos2[X] = pos1[X];pos1[X] = tmp;}
  if (pos1[Y]>pos2[Y]){tmp = pos2[Y];pos2[Y] = pos1[Y];pos1[Y] = tmp;}

  // calculate size & pos
  st->size[X] = pos2[X]-pos1[X];
  st->size[Y] = pos2[Y]-pos1[Y];
  st->size[Z] = toFIXED(1.0);

  st->pos[X] = pos1[X]+slDivFX(toFIXED(2.0),st->size[X]);
  st->pos[Y] = pos1[Y]+slDivFX(toFIXED(2.0),st->size[Y]);
  st->pos[Z] = INFINITE;

  // st->lifetime = minmaxrnd(toFIXED(200.0),toFIXED(2000.0));
}

void station_display(STATION* st){

  if (st->state == away) return;  

  switch (st->state){
  case coming :  
    slPushMatrix();
    {
      slTranslate(st->pos[X],st->pos[Y],st->pos[Z]);
      slPushMatrix();
      {
	slScale(st->size[X],st->size[Y],st->size[Z]);
	slPushMatrix();
	{
	  st->ang[Z] += DEGtoANG(20.0); 
	  slRotZ(st->ang[Z]);
	  slPutPolygon(st->gfx);      
	}
	slPopMatrix();
      }
      slPopMatrix();
    }
    slPopMatrix();
    break;

  case here : 
    {
      int i;
      FIXED zpos = st->pos[Z];
      for (i=0;i<1;i++){
	//	zpos += slMulFX(slRandom(),toFIXED(S_SIZE/20.0));
	slPushMatrix();
	{
	  slTranslate(st->pos[X],st->pos[Y],zpos);
	  slPushMatrix();
	  {
	    slScale(st->size[X],st->size[Y],st->size[Z]);
	    if (st->pos[Z] > S_ZM){ // .. or we won't see the players anymore
	      slPutPolygon(st->gfx);      
	    } else {
	      slPutPolygon(st->gfx_trans);
	      slPushMatrix();
	      {
		slScale(toFIXED(0.5),toFIXED(0.5),toFIXED(1.0));
		slPutPolygon(st->gfx_trans);
		slPushMatrix();
		{
		  slScale(toFIXED(0.4),toFIXED(0.4),toFIXED(1.0));
		  slPutPolygon(st->gfx_trans);
		}
		slPopMatrix();
	      }
	      slPopMatrix();
	    }
	  }
	  slPopMatrix();
	}
	slPopMatrix();
      }
    }

    break;

  case leaving :
    st->size[X] = slMulFX(toFIXED(0.98),st->size[X]);
    st->size[Y] = slMulFX(toFIXED(0.98),st->size[Y]);
    //st->size[Z] = slMulFX(toFIXED(1.0),st->size[Z]);
    slPushMatrix();
    {
      slTranslate(st->pos[X],st->pos[Y],st->pos[Z]);
      slPushMatrix();
      {
	slScale(st->size[X],st->size[Y],st->size[Z]);
	slPushMatrix();
	{
	  st->ang[Y] += DEGtoANG(20.0); 
	  slRotY(st->ang[Y]);
	  slPutPolygon(st->gfx_trans); 
	}
	slPopMatrix();
      }
      slPopMatrix();
    }
    slPopMatrix();

  default:
    break;
  }
}

void station_leave(STATION* st){
  st->r_time =  minmaxrnd(toFIXED(25.0),toFIXED(45.0));
  st->state = leaving;
}

void station_step_sm(STATION* st){
  FIXED rt = st->r_time;
  station_state ss = st->state;

  st->pos[Z] -= st->zspeed;

  switch (ss){
  case away : 
    if (rt < toFIXED(0.0)) {
      station_randomize(st);
      rt = minmaxrnd(toFIXED(25.0),toFIXED(50.0));
      ss = coming;
    }
    break ;
  case coming :  
    if (rt < toFIXED(0.0)) {
      // rt = st->lifetime;
      ss = here;
    }
    break ; 
  case here :  
    //    if (rt < toFIXED(0.0)) {
    if ( st->pos[Z] < slMulFX(toFIXED(2.0),S_Zm)) {
      rt = minmaxrnd(toFIXED(10.0),toFIXED(30.0));
      ss = leaving;
    }
    break; 
  case leaving : 
    if (rt < toFIXED(0.0)) {
      //      rt = minmaxrnd(toFIXED(200.0),toFIXED(10000.0));
      rt = toFIXED(1.0); // direct -> away
      ss = away;
    }
    break;
  default : break;
  }

  rt -= toFIXED(1.0);

  st->r_time = rt;
  st->state = ss;
}

/*
 * TAIL (of a particle)
 */


POINT point_TAIL[]= {
  POStoFIXED(-(S_SIZE/40.0),-(S_SIZE/30.0), 0.0),  //0 
  POStoFIXED( (S_SIZE/40.0),-(S_SIZE/40.0), 0.0),  //1
  POStoFIXED( 0.0,-1.0, 0.0),    //2
  POStoFIXED(-(S_SIZE/40.0),-(S_SIZE/40.0), 0.0),  //3=0
};

POLYGON polygon_TAIL[]= {
  NORMAL(0.0,1.0,0.0),
  VERTICES(0,1,2,3),
};

ATTR attribute_TAIL1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(14,1,1),No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_TAIL2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(14,14,1),No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_TAIL3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(1,14,1),No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_TAIL4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(1,1,14),No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_TAIL5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(1,25,25),No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

PDATA PD_TAIL1 = {
  point_TAIL,sizeof(point_TAIL)/sizeof(POINT),
  polygon_TAIL,sizeof(polygon_TAIL)/sizeof(POLYGON),
  attribute_TAIL1
};
PDATA PD_TAIL2 = {
  point_TAIL,sizeof(point_TAIL)/sizeof(POINT),
  polygon_TAIL,sizeof(polygon_TAIL)/sizeof(POLYGON),
  attribute_TAIL2
};
PDATA PD_TAIL3 = {
  point_TAIL,sizeof(point_TAIL)/sizeof(POINT),
  polygon_TAIL,sizeof(polygon_TAIL)/sizeof(POLYGON),
  attribute_TAIL3
};
PDATA PD_TAIL4 = {
  point_TAIL,sizeof(point_TAIL)/sizeof(POINT),
  polygon_TAIL,sizeof(polygon_TAIL)/sizeof(POLYGON),
  attribute_TAIL4
};
PDATA PD_TAIL5 = {
  point_TAIL,sizeof(point_TAIL)/sizeof(POINT),
  polygon_TAIL,sizeof(polygon_TAIL)/sizeof(POLYGON),
  attribute_TAIL5
};

#define MAX_TAIL_SIZE 30

typedef enum {on,off} ONOFF;

typedef struct _TAIL{
  FIXED pos[MAX_TAIL_SIZE][XYZ];
  int first;
  int last;
  ONOFF wo;
  PDATA* gfx[MAX_TAIL_SIZE];
} TAIL;

// returns last tail part indix or -1 if n/a
int tail_get_last(TAIL* ta) {
  int idx;
  if (ta->wo == on){
    idx = ta->last;
  } else {
    idx = -1;
  }
  return idx;

}

void tail_init (TAIL* ta){ // !! MAX_TAIL_SIZE when changing it
  int i=0;

  ta->first = 0;
  ta->last  = 0;
  ta->wo = off;

  // pas beau
  /*  for (i=0;((i<2)&&(i<MAX_TAIL_SIZE));i++){
      ta->gfx[i] = &PD_TAIL1;
      }
      for (i=2;((i<5)&&(i<MAX_TAIL_SIZE));i++){
      ta->gfx[i] = &PD_TAIL2;
      }
      for (i=6;((i<11)&&(i<MAX_TAIL_SIZE));i++){
      ta->gfx[i] = &PD_TAIL3;
      }
      for (i=12;((i<18)&&(i<MAX_TAIL_SIZE));i++){
      ta->gfx[i] = &PD_TAIL4;
      }
      for (i=19;i<MAX_TAIL_SIZE;i++){
      ta->gfx[i] = &PD_TAIL5;
      }
  */

  // pas beau non plus.
  for (i=0;i<MAX_TAIL_SIZE;i++){
    FIXED rnd = slRandom();
    if (rnd>toFIXED(0.8)) {ta->gfx[i] = &PD_TAIL5;} else {
      if (rnd>toFIXED(0.6)) {ta->gfx[i] = &PD_TAIL4;} else {
	if (rnd>toFIXED(0.4)) {ta->gfx[i] = &PD_TAIL3;} else {
	  if (rnd>toFIXED(0.2)) {ta->gfx[i] = &PD_TAIL2;} else {
	    ta->gfx[i] = &PD_TAIL1;
	  }
	}
      }
    }
  }

}

// adds a x,y,z pos to the tail (replaces if full)
void tail_update (TAIL* ta, FIXED x, FIXED y, FIXED z){
  int last, first;

  last = ta->last;
  first = ta->first;

  if (last == first) { // only for the first update
    ta->wo = on;
  } else {
    last++;
  }

  if (last == MAX_TAIL_SIZE) {
    last = 0;
  }
  if (last == first) {
    first ++;
    if (first == MAX_TAIL_SIZE){
      first = 0;
    }
  }
  ta->pos[last][X]=x; 
  ta->pos[last][Y]=y; 
  ta->pos[last][Z]=z;
  ta->first = first;
  ta->last  = last;
}


void tail_apply_wind(TAIL* ta,FIXED* wi){
  int i;
  for (i=0;i<MAX_TAIL_SIZE;i++){//brutal
    ta->pos[i][X]+=wi[X]; 
    ta->pos[i][Y]+=wi[Y]; 
    ta->pos[i][Z]+=wi[Z];
  }
}

void tail_display(TAIL* ta){
  int first, last;
  int i,cur,count;

  cur = first = ta->first;
  last = ta->last;

  if (ta->wo == off) return; // no tail.

  count = (last>=first)?(last-first+1):(MAX_TAIL_SIZE-first+last);
  for (i=0;i<count;i++){
    // display
    slPushMatrix();
    {
      slTranslate(ta->pos[cur][X],ta->pos[cur][Y],ta->pos[cur][Z]);
      slPushMatrix();
      {
	/*
	  slRotZ(slRandom());
	  slRotX(slRandom());
	  slRotY(slRandom());
	*/
	slPutPolygon(ta->gfx[MAX_TAIL_SIZE-i-1]);    
      }
      slPopMatrix();
    }
    slPopMatrix();
    cur++;
    if (cur==MAX_TAIL_SIZE) cur = 0;
  }
}

/*
 * LIFE BAR
 */


static POINT point_LIFE_BAR[]={
  POStoFIXED(0.0, 0.8,0.0),
  POStoFIXED(1.0,0.8,0.0),
  POStoFIXED(1.0,-0.8,0.0),
  POStoFIXED(0.0,-0.8,0.0),
};

POLYGON polygon_LIFE_BAR[]= {
  NORMAL(0.0,0.0,1.0),
  VERTICES(0,1,2,3),
};

ATTR attribute_LIFE_BAR1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };
ATTR attribute_LIFE_BAR2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };
ATTR attribute_LIFE_BAR3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };
ATTR attribute_LIFE_BAR4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };
ATTR attribute_LIFE_BAR5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };
ATTR attribute_LIFE_BAR6[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_p,No_Gouraud,MESHoff,sprPolygon,UseLight), };

PDATA PD_LIFE_BAR1 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR1
};
PDATA PD_LIFE_BAR2 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR2
};
PDATA PD_LIFE_BAR3 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR3
};
PDATA PD_LIFE_BAR4 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR4
};
PDATA PD_LIFE_BAR5 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR5
};
PDATA PD_LIFE_BAR6 = {
  point_LIFE_BAR,sizeof(point_LIFE_BAR)/sizeof(POINT),
  polygon_LIFE_BAR,sizeof(polygon_LIFE_BAR)/sizeof(POLYGON),
  attribute_LIFE_BAR6
};

void life_bar_display(int nplayer, FIXED energy, int total){

  FIXED Xl,Xpos,Xlenght;
  FIXED max_bar_lenght;

  Xl = S_XM - S_Xm;
  max_bar_lenght = slDivFX(toFIXED(1.0*(total+1)),Xl);
  Xpos = S_Xm + slMulFX(toFIXED(1.0*nplayer),max_bar_lenght);
  Xlenght = slMulFX(max_bar_lenght,slDivFX(MAX_PLAYER_ENERGY,energy));

  slPushMatrix(); 
  {
    slTranslate(Xpos,S_Ym,slDivFX(toFIXED(2.0),S_Zm));
    slPushMatrix(); 
    {
      slScale(Xlenght,toFIXED(1.0),toFIXED(1.0));
      switch(nplayer){
      case 0 :  slPutPolygon(&PD_LIFE_BAR1); break;
      case 1 :  slPutPolygon(&PD_LIFE_BAR2); break;
      case 2 :  slPutPolygon(&PD_LIFE_BAR3); break;
      case 3 :  slPutPolygon(&PD_LIFE_BAR4); break;
      case 4 :  slPutPolygon(&PD_LIFE_BAR5); break;
      case 5 :  slPutPolygon(&PD_LIFE_BAR6); break;
      default :  
	slPutPolygon(&PD_LIFE_BAR1);
      }
    }
    slPopMatrix();
  }
  slPopMatrix();

}


/*
 * LOCALIZATION PLAN (for a particle)
 */

static POINT point_LOC_PLAN[]={
  POStoFIXED(-S_SIZE, -S_SIZE/100.0,0.0),
  POStoFIXED( S_SIZE, -S_SIZE/100.0,0.0),
  POStoFIXED( S_SIZE,  S_SIZE/100.0,0.0),
  POStoFIXED(-S_SIZE,  S_SIZE/100.0,0.0),
};

POLYGON polygon_LOC_PLAN[]= {
  NORMAL(0.0,0.0,1.0),
  VERTICES(0,1,2,3),
};

ATTR attribute_LOC_PLAN1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };
ATTR attribute_LOC_PLAN2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };
ATTR attribute_LOC_PLAN3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };
ATTR attribute_LOC_PLAN4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };
ATTR attribute_LOC_PLAN5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };
ATTR attribute_LOC_PLAN6[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_p,No_Gouraud,MESHoff,sprPolyLine,UseLight), };

PDATA PD_LOC_PLAN1 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN1
};
PDATA PD_LOC_PLAN2 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN2
};
PDATA PD_LOC_PLAN3 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN3
};
PDATA PD_LOC_PLAN4 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN4
};
PDATA PD_LOC_PLAN5 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN5
};
PDATA PD_LOC_PLAN6 = {
  point_LOC_PLAN,sizeof(point_LOC_PLAN)/sizeof(POINT),
  polygon_LOC_PLAN,sizeof(polygon_LOC_PLAN)/sizeof(POLYGON),
  attribute_LOC_PLAN6
};

void localization_display(PDATA* pd_loc, FIXED* pos){

  slPushMatrix(); // Y
  {
    slTranslate(toFIXED(0.0),pos[Y],pos[Z]);
    slPutPolygon(pd_loc);
  }
  slPopMatrix();

  slPushMatrix(); // Y
  {
    slTranslate(pos[X],toFIXED(0.0),pos[Z]);
    slPushMatrix(); 
    {
      slRotZ(DEGtoANG(90.0));
      slPutPolygon(pd_loc);
    }
    slPopMatrix();	
  }
  slPopMatrix();
}


/*
 * PARTICLES
 */

POINT point_PAR[]= {
  POStoFIXED(-.0,-(S_SIZE/10.0), 0.0),  //0 
  POStoFIXED( (S_SIZE/10.0),-(S_SIZE/10.0), 0.0),  //1
  POStoFIXED( 0.0,-(S_SIZE/50.0), 0.0),  //2
  POStoFIXED(-(S_SIZE/10.0),-(S_SIZE/10.0), 0.0),  //3=0
  POStoFIXED(-(S_SIZE/10.0), (S_SIZE/10.0), 0.0),  //4
  POStoFIXED( (S_SIZE/10.0), (S_SIZE/10.0), 0.0),  //5
  POStoFIXED( 0.0, (S_SIZE/50.0), 0.0),  //6
  POStoFIXED(-(S_SIZE/10.0), (S_SIZE/10.0), 0.0),  //7
};

POLYGON polygon_PAR[]= {
  NORMAL(0.0,1.0,0.0),
  VERTICES(0,1,2,3),
  NORMAL(0.0,1.0,0.0),
  VERTICES(4,5,6,7),
};

ATTR attribute_PAR1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_1_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PAR2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_2_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PAR3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_3_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PAR4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_4_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PAR5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_5_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PAR6[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_6_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

ATTR attribute_PARn[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_n_p,No_Gouraud,MESHoff,sprPolygon,UseLight),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,CO_n_s,No_Gouraud,MESHoff,sprPolygon,UseLight),  
};

PDATA PD_PAR1 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR1
};

PDATA PD_PAR2 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR2
};

PDATA PD_PAR3 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR3
};

PDATA PD_PAR4 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR4
};

PDATA PD_PAR5 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR5
};
PDATA PD_PAR6 = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PAR6
};
PDATA PD_PARn = {
  point_PAR,sizeof(point_PAR)/sizeof(POINT),
  polygon_PAR,sizeof(polygon_PAR)/sizeof(POLYGON),
  attribute_PARn
};

// three types of charge
typedef enum _CHARGE_TYPE{ct_1=0,ct_2,ct_3} CHARGE_TYPE;
typedef enum _PA_STYLE{pas_player, pas_attractor} PA_STYLE;

typedef struct _PARTICLE {
  FIXED e_charge[3];       // e_charge[charge_type]
  int charged[3];          // charged[charge_type]
  int can_move;
  FIXED pos[XYZ]; 
  ANGLE ang[XYZ];
  FIXED speed[XYZ];
  TAIL tail;
  int display_tail;
  int display_loc;
  int display;
  XPDATA*  gfx;
  PDATA*  loc_gfx;
  PA_STYLE  style;
} PARTICLE ;

void particle_init (PARTICLE* pa,XPDATA* gfx,PDATA* loc_gfx, PA_STYLE style){
  pa->pos[X] = minmaxrnd(S_Xm,S_XM);
  pa->pos[Y] = minmaxrnd(S_Ym,S_YM);
  pa->pos[Z] = minmaxrnd(S_Zm,S_ZM);
  pa->speed[X] = toFIXED(0.0);// slRandom();
  pa->speed[Y] = toFIXED(0.0);// slRandom();
  pa->speed[Z] = toFIXED(0.0);// slRandom();
  pa->ang[X] = pa->ang[Y] = pa->ang[Z] = DEGtoANG(0.0);
  pa->e_charge[ct_1] = 0; pa->e_charge[ct_2] = 0; pa->e_charge[ct_3] = 0;
  pa->can_move = TRUE;
  pa->charged[ct_1] = FALSE; pa->charged[ct_2] = FALSE; pa->charged[ct_3] = FALSE; 
  pa->display_tail = TRUE;
  pa->gfx = gfx;
  if (loc_gfx == (PDATA*)NULL) {
    pa->loc_gfx = (PDATA*)NULL;
    pa->display_loc = FALSE;
  } else {
    pa->loc_gfx = loc_gfx;
    pa->display_loc = FALSE;
  }
  pa->display = TRUE;
  tail_init(&(pa->tail));
  pa->style = style;
}

void particle_randomize_pos(PARTICLE* pa){
  pa->pos[X] = minmaxrnd (S_Xm, S_XM);
  pa->pos[Y] = minmaxrnd (S_Ym, S_YM);
  pa->pos[X] = minmaxrnd (S_Zm, S_ZM);
}

void particle_set_charge(PARTICLE* pa,CHARGE_TYPE ct,FIXED c){
  pa->e_charge[ct] = c;
}

void particle_move (PARTICLE* pa){
  // move
  FIXED mms = slMulFX(MAX_SPEED,toFIXED(-1.0));

  if (pa->speed[X] > MAX_SPEED) pa->speed[X] = MAX_SPEED;
  if (pa->speed[Y] > MAX_SPEED) pa->speed[Y] = MAX_SPEED;
  if (pa->speed[Z] > MAX_SPEED) pa->speed[Z] = MAX_SPEED;
  if (pa->speed[X] < mms) pa->speed[X] = mms;
  if (pa->speed[Y] < mms) pa->speed[Y] = mms;
  if (pa->speed[Z] < mms) pa->speed[Z] = mms;

  pa->pos[X] += pa->speed[X];
  pa->pos[Y] += pa->speed[Y];
  pa->pos[Z] += pa->speed[Z];

  // collision with screen limits
  if ((pa->pos[X]>S_XM) || (pa->pos[X]<S_Xm)) {
    sound_play(snd_bing);
    pa->speed[X] = slMulFX(pa->speed[X],toFIXED(-1.5));
  }
  if ((pa->pos[Y]>S_YM) || (pa->pos[Y]<S_Ym)) {
    sound_play(snd_bing);
    pa->speed[Y] = slMulFX(pa->speed[Y],toFIXED(-1.5));
  }
  if ((pa->pos[Z]>S_ZM) || (pa->pos[Z]<S_Zm)) {
    sound_play(snd_bing);
    pa->speed[Z] = slMulFX(pa->speed[Z],toFIXED(-1.5));
  }

  
  if (pa->pos[X]>S_XM) pa->pos[X] = S_XM;// - slRandom();
  if (pa->pos[X]<S_Xm) pa->pos[X] = S_Xm;// + slRandom();
  if (pa->pos[Y]>S_YM) pa->pos[Y] = S_YM;// - slRandom();
  if (pa->pos[Y]<S_Ym) pa->pos[Y] = S_Ym;// + slRandom();
  if (pa->pos[Z]>S_ZM) pa->pos[Z] = S_ZM;// - slRandom();
  if (pa->pos[Z]<S_Zm) pa->pos[Z] = S_Zm;// + slRandom();
  
  /* // break on through
    if (pa->pos[X]>S_XM) pa->pos[X] = S_Xm;// - slRandom();
    if (pa->pos[X]<S_Xm) pa->pos[X] = S_XM;// + slRandom();
    if (pa->pos[Y]>S_YM) pa->pos[Y] = S_Ym;// - slRandom();
    if (pa->pos[Y]<S_Ym) pa->pos[Y] = S_YM;// + slRandom();
    if (pa->pos[Z]>S_ZM) pa->pos[Z] = S_Zm;// - slRandom();
    if (pa->pos[Z]<S_Zm) pa->pos[Z] = S_ZM;// + slRandom();
  */
  tail_update((&(pa->tail)),pa->pos[X],pa->pos[Y],pa->pos[Z]);
}


void particle_set_can_move(PARTICLE* pa,int canmove){
  pa->can_move = canmove;
}

void particle_set_is_charged(PARTICLE* pa,CHARGE_TYPE ct,int charged){
  pa->charged[ct] = charged;
}

void particle_invertCharge(PARTICLE* pa,CHARGE_TYPE ct){
  particle_set_charge(pa,ct,slMulFX(toFIXED(-1.0),pa->e_charge[ct]));
}
void particle_switch_display_loc(PARTICLE* pa){
  pa->display_loc = (pa->display_loc)?FALSE:TRUE;
}

void particle_display(PARTICLE* pa){ 

  static ANGLE rot_value[] = {
    DEGtoANG(18.0),DEGtoANG(15.0),DEGtoANG(9.0),
    DEGtoANG(-9.0),DEGtoANG(-15.0),DEGtoANG(-18.0),
  };

  if (pa->display_tail)
    tail_display(&(pa->tail));    

  if (pa->display_loc)
    localization_display(pa->loc_gfx,pa->pos);  

  if (pa->display){
    switch (pa->style) {
    case pas_player :
      slPushMatrix();
      {    
	FIXED c = pa->e_charge[ct_1]+pa->e_charge[ct_2]+pa->e_charge[ct_3];   
	ANGLE a = pa->ang[Z];
	slTranslate(pa->pos[X],pa->pos[Y],pa->pos[Z]);
	slPushMatrix();
	{
	  if (c >= toFIXED(2.0)) a = a + rot_value[0];
	  else if (c >= toFIXED( 1.0)) a = a + rot_value[1];
	  else if (c > toFIXED( 0.0)) a = a + rot_value[2];
	  else if (c <= toFIXED( -2.0)) a = a + rot_value[3];
	  else if (c <= toFIXED(-1.0)) a = a + rot_value[4];
	  else if (c < toFIXED(-0.0)) a = a + rot_value[5];
	  pa->ang[Z] = a;
	  slRotZ(a);
	  slPutPolygon((PDATA*)pa->gfx); 
	}
	slPopMatrix(); 
      }
      slPopMatrix();
      break;
    case pas_attractor :
      slPushMatrix();
      {    
	FIXED light[] = {toFIXED(1.1),toFIXED(0.2),toFIXED(0.3)};

	slTranslate(pa->pos[X],pa->pos[Y],pa->pos[Z]);
	slPushMatrix();
	{
	  slScale(toFIXED(3.0),toFIXED(3.0),toFIXED(3.0));
	  slPushMatrix();
	  {
	    ANGLE* a = pa->ang;
	    if (pa->charged[ct_1]) a[X] += 0x0900;
	    if (pa->charged[ct_2]) a[Y] += 0x0800;
	    if (pa->charged[ct_3]) a[Z] += 0x0700;
	    slRotX(a[X]);slRotY(a[Y]);slRotZ(a[Z]);
	    slPutPolygonX(pa->gfx,light); 
	  }  
	  slPopMatrix();
	}
	slPopMatrix(); 
      }
      slPopMatrix();
      break;
    default : break;
    }
  }
}

// slow down <pa> according to the rugosity
void particle_applyRugosity(PARTICLE* pa, FIXED ru){
  pa->speed[X] = slMulFX(pa->speed[X],ru);
  pa->speed[Y] = slMulFX(pa->speed[Y],ru);
  pa->speed[Z] = slMulFX(pa->speed[Z],ru);
}

// slow down <pa> according to the rugosity
void particle_applyGravity(PARTICLE* pa, FIXED gr){
  if (pa->can_move)
    pa->speed[Y] += gr;
}


/*
 * PLAYERS
 */
typedef enum _player_state {pl_coming, pl_here, pl_leaving, pl_dead, pl_away} player_state;
typedef struct _PLAYER {
  FIXED energy;
  int points;
  int pad_num;
  Uint16 last_pad;
  player_state state;
  PARTICLE* particle;  // particule owned by the player
  FIXED r_time;
  int wants_to_come;
  int wants_to_leave;
} PLAYER;

void player_init(PLAYER* p, int pad_num, PARTICLE* par){
  p->points = 0;
  p->pad_num = pad_num;
  p->last_pad = Smpc_Peripheral[pad_num].data;
  p->energy = toFIXED(10.0);

  p->state = pl_away;
  p->wants_to_come = FALSE;
  p->wants_to_leave = FALSE;
  p->r_time = minmaxrnd(toFIXED(0.0),toFIXED(10.0)); // minimal time away

  p->particle = par;
  {
    par->display_tail = FALSE;
    par->display_loc = FALSE;
    par->display = FALSE;
    par->charged[ct_1] = FALSE;par->charged[ct_2] = FALSE;par->charged[ct_3] = FALSE;
  }
}
int player_is_here (PLAYER* p){
  return (p->state == pl_here);
}
void player_hurt(PLAYER* p){
  p->energy -= toFIXED(1.0);
}
void player_do_some_good(PLAYER* p){
  p->energy += toFIXED(1.0); 
  if (p->energy > MAX_PLAYER_ENERGY) p->energy = MAX_PLAYER_ENERGY;
}

void player_step_sm(PLAYER* p){

  PARTICLE* par = p->particle;
  player_state st = p->state;
  FIXED r_time =  p->r_time;

  switch (st){
  case pl_coming : 
    r_time -= toFIXED(1.0);
    if (r_time <= toFIXED(0.0)){
      p->wants_to_come = FALSE;
      par->display_tail = TRUE;
      par->display = TRUE;
      par->can_move = TRUE;
      par->display_loc = FALSE;
      st = pl_here;     
      //      scrollTxtAddTxtLine("here");
    }
    break;
  case pl_here : 
    if ((p->wants_to_leave) || (p->energy<=toFIXED(0.0))){
      p->wants_to_leave = FALSE;
      par->display_tail = TRUE;
      par->display = FALSE;
      par->display_loc = FALSE;
      r_time = minmaxrnd(toFIXED(50.0),toFIXED(100.0));  // leaving time
      st = pl_leaving;
      sound_play(snd_death);
    } 
    break;
  case pl_leaving : 
    r_time = r_time - toFIXED(1.0);
    if (r_time <= toFIXED(0.0)){
      par->display_tail = FALSE;
      par->display = FALSE;
      par->charged[ct_1] = FALSE; par->charged[ct_2] = FALSE; par->charged[ct_3] = FALSE;
      par->can_move = FALSE;
      r_time = minmaxrnd(toFIXED(10.0),toFIXED(25.0)); // minimal time away
      st = pl_away;
    } 
    break;
  case pl_away : 
    if (p->wants_to_come){
      r_time = r_time - toFIXED(1.0); // tait au dessus du p->wantsto...
      if (r_time<toFIXED(0.0)){
	p->wants_to_come = FALSE;
	p->energy = toFIXED(10.0);
	par->display_tail = FALSE;
	par->display = TRUE;
	par->display_loc = TRUE;
	r_time = minmaxrnd(toFIXED(50.0),toFIXED(100.0)); // coming time
	st = pl_coming;
	//scrollTxtAddTxtLine("hello           ");
	// scrollTxtAddTxtLine("coming");
      }
    }
    break;
  default: break;
  }
  p->r_time = r_time;
  p->state = st;
}

/*
 * GAME Params
 */

typedef struct _GAME_PARAMS {
  int display;
  FIXED rugosity;    // weight of last synch()'s particle speed (1:no rugosity;
		     // 0:last speed not considered)
  FIXED charge_factor; 
  FIXED gravity;
  FIXED spm_pseudo_speed;
  FIXED spm_size_z;
  FIXED camera_dflt_src_pos[XYZ];
  FIXED camera_dflt_dst_pos[XYZ];
  FIXED camera_dflt_intro_src_pos[XYZ];
  FIXED camera_dflt_intro_dst_pos[XYZ];
  int n_players;     // number of player
  int cur_set;       // cur set number
  int n_set;         // set number
} GAME_PARAMS;

 void game_params_init(GAME_PARAMS* gp){
   gp->display = FALSE;
   gp->rugosity = toFIXED(0.94); //
   gp->charge_factor = toFIXED(0.70);
   gp->gravity = toFIXED(0.12);
   gp->camera_dflt_src_pos[X] = toFIXED(0.0);
   gp->camera_dflt_src_pos[Y] = slMulFX(toFIXED(0.8),S_Ym);
   gp->camera_dflt_src_pos[Z] = slMulFX(toFIXED(2.2),S_Zm);
   gp->camera_dflt_dst_pos[X] = toFIXED(0.0);
   gp->camera_dflt_dst_pos[Y] = toFIXED(0.0);
   gp->camera_dflt_dst_pos[Z] = toFIXED(0.0);

   gp->camera_dflt_intro_src_pos[X] = slMulFX(toFIXED(10.0),minmaxrnd(S_Xm,S_XM));
   gp->camera_dflt_intro_src_pos[Y] = slMulFX(toFIXED(10.0),minmaxrnd(S_Ym,S_YM));
   gp->camera_dflt_intro_src_pos[Z] = slMulFX(toFIXED(10.0),minmaxrnd(S_Zm,S_ZM));

   gp->camera_dflt_intro_dst_pos[X] = toFIXED(0.0);
   gp->camera_dflt_intro_dst_pos[Y] = toFIXED(0.0);
   gp->camera_dflt_intro_dst_pos[Z] = toFIXED(0.0);

   gp->spm_pseudo_speed = toFIXED(S_SIZE/90.0);
   gp->spm_size_z = toFIXED((2.0*S_SIZE)/9.0);
   gp->n_players = 0;
   gp->cur_set = 0;
   gp->n_set = 3;
 }


 void game_params_print(GAME_PARAMS* gp){
   int i = 18;
   if (gp->display) {
     // slCurColor(7);
     slPrintFX(gp->rugosity,slLocate(0,i++));
     //  slCurColor(4);
     slPrintFX(gp->charge_factor,slLocate(0,i++));
     //  slCurColor(10);
     slPrintFX(gp->gravity,slLocate(0,i++));
     slPrintFX(gp->spm_pseudo_speed,slLocate(0,i++));
     slPrintFX(gp->spm_size_z,slLocate(0,i++));
   } else {
     slPrint("                    ",slLocate(0,i++));
     slPrint("                    ",slLocate(0,i++));
     slPrint("                    ",slLocate(0,i++));
     slPrint("                    ",slLocate(0,i++));
     slPrint("                    ",slLocate(0,i++));
   }
 }

 /*
  * CAMERA
  */
 typedef enum _cam_type {external, internal} CAM_TYPE;

 typedef struct _CAMERA {
   CAM_TYPE  src_type;      
   CAM_TYPE  dst_type;

   FIXED*    ext_src_pos;     // target pos (ext or int)
   FIXED*    ext_dst_pos;
   FIXED     int_src_pos[XYZ];  
   FIXED     int_dst_pos[XYZ];

   FIXED     c_src_pos[XYZ];  // current pos
   FIXED     c_dst_pos[XYZ];
   FIXED     time; // moving time
 } CAMERA;

 void camera_init(CAMERA* ca, FIXED* src, FIXED* dst){
   ca->src_type = internal;
   ca->dst_type = internal;
   ca->ext_src_pos = (FIXED*) NULL;
   ca->ext_dst_pos = (FIXED*) NULL;
   ca->int_src_pos[X] = src[X];
   ca->int_src_pos[Y] = src[Y];
   ca->int_src_pos[Z] = src[Z];
   ca->int_dst_pos[X] = dst[X];
   ca->int_dst_pos[Y] = dst[Y];
   ca->int_dst_pos[Z] = dst[Z];
   ca->time = toFIXED(0.0);
   ca->c_src_pos[X] = toFIXED(0.0); // ca->int_src_pos[X];
   ca->c_src_pos[Y] = toFIXED(0.0); // ca->int_src_pos[Y];
   ca->c_src_pos[Z] = toFIXED(0.0); // ca->int_src_pos[Z];
   ca->c_dst_pos[X] = toFIXED(0.0); // ca->int_dst_pos[Y];
   ca->c_dst_pos[Y] = toFIXED(0.0); // ca->int_dst_pos[Y]
   ca->c_dst_pos[Z] = toFIXED(0.0); // ca->int_dst_pos[Z];
 }

 void camera_setMoveTime(CAMERA* ca, FIXED time){ca->time = time ;}

 void camera_step(CAMERA* ca){
   FIXED* src,*dst;

   ca->time -= toFIXED(1.0);
   if (ca->time < toFIXED(0.0)){
     ca->time = toFIXED(0.0);
   }

   // source
   switch (ca->src_type){
   case internal : src = (FIXED*) ca->int_src_pos; break;
   case external : src = ca->ext_src_pos; break;
   default : src = (FIXED*) NULL;
   }
   if (ca->time > 0){
     int i;
     for (i=0;i<XYZ;i++){
       FIXED c,f;
       c = ca->c_src_pos[i];
       f = src[i];
       c += slDivFX(ca->time,f-c);
       ca->c_src_pos[i] = c;
     }
   } else {
     int i;
     for (i=0;i<XYZ;i++){
       ca->c_src_pos[i] = src[i];
     }
   }

   // dest 
   switch (ca->dst_type){
   case internal : dst = (FIXED*) ca->int_dst_pos; break;
   case external : dst = ca->ext_dst_pos; break;
   default : dst = (FIXED*) NULL;
   }
   if (ca->time > 0){
     int i;
     for (i=0;i<XYZ;i++){
       FIXED c,f;
       c = ca->c_dst_pos[i];
       f = dst[i];
       c += slDivFX(ca->time,f-c);
       ca->c_dst_pos[i] = c;
     }
   } else {
     int i;
     for (i=0;i<XYZ;i++){
       ca->c_dst_pos[i] = dst[i];
     }
   }
 }

 void camera_useInternalSource(CAMERA* ca){ca->src_type = internal;}
 void camera_useExternalSource(CAMERA* ca){ca->src_type = external;}
 void camera_useInternalDest  (CAMERA* ca){ca->dst_type = internal;}
 void camera_useExternalDest  (CAMERA* ca){ca->dst_type = external;}

 void camera_setExternalSource(CAMERA* ca,FIXED* pos){ca->ext_src_pos = pos;}
 void camera_setExternalDest  (CAMERA* ca,FIXED* pos){ca->ext_dst_pos = pos;}

 void camera_setInternalSource(CAMERA* ca,FIXED* pos){
   ca->int_src_pos[X] = pos[X];
   ca->int_src_pos[Y] = pos[Y];
   ca->int_src_pos[Z] = pos[Z];
 }
 void camera_setInternalDest(CAMERA* ca,FIXED* pos){
   ca->int_dst_pos[X] = pos[X];
   ca->int_dst_pos[Y] = pos[Y];
   ca->int_dst_pos[Z] = pos[Z];
 }

 FIXED* camera_getInternalSource(CAMERA* ca){return (ca->int_src_pos);}
 FIXED* camera_getInternalDest(CAMERA* ca)  {return (ca->int_dst_pos);}
 FIXED* camera_getSource(CAMERA* ca){  return ca->c_src_pos;}
 FIXED* camera_getDest(CAMERA* ca){  return ca->c_dst_pos;}

void camera_oscillate(CAMERA* ca){
  if (ca->time <= toFIXED(0.0)){
    FIXED* p = ca->int_src_pos ;
    p[X] = minmaxrnd(toFIXED(-S_SIZE/20.0),toFIXED(S_SIZE/20.0));
    p[Y] = minmaxrnd(toFIXED(-S_SIZE/15.0),toFIXED(S_SIZE/15.0));
    ca->time = minmaxrnd(toFIXED(20.0),toFIXED(40.0));
   }
}

/*
 * LIGHT
 */

typedef struct _LIGHT {
  FIXED v[XYZ];
} LIGHT;

void light_init(LIGHT* l){
  l->v[X] = toFIXED(0.0);
  l->v[Y] = toFIXED(0.0);
  l->v[Z] = toFIXED(1.0);
}


void light_randomize(LIGHT* l){
  l->v[X] = minmaxrnd(toFIXED(0.0),toFIXED(1.0));
  l->v[Y] = minmaxrnd(toFIXED(0.0),toFIXED(1.0));
  l->v[Z] = minmaxrnd(toFIXED(0.0),toFIXED(1.0));
}


void light_move(LIGHT* l){
  l->v[X] += minmaxrnd(toFIXED(-0.05),toFIXED(0.05));
  l->v[Y] += minmaxrnd(toFIXED(-0.05),toFIXED(0.05));
  l->v[Z] += minmaxrnd(toFIXED(-0.05),toFIXED(0.05)); 
}


 /*
  * GAME
  */

 typedef enum _GAME_STEP {g_splash, g_start, g_intro, g_in_match, g_outro, g_end_match} GAME_STEP;
 typedef enum _GAME_COMMAND {gc_next, gc_end_match, gc_end_game} GAME_COMMAND;

 typedef struct _GAME {
   GAME_STEP step;
 } GAME;

 void game_init (GAME* ga){
   ga->step = g_splash;
 }

 void game_set_step(GAME* ga, GAME_COMMAND cmd){

   GAME_STEP st = ga->step;

   switch (st){
   case g_splash : st = g_start;
     break;
   case g_start  : st = g_intro;
     //     scrollTxtAddTxtLine("g_intro");
     break;
   case g_intro  : st = g_in_match;
     //scrollTxtAddTxtLine("in_match");
     break;
   case g_in_match : st = g_outro;
     //scrollTxtAddTxtLine("g_outro");
     break;
   case g_outro  :
     switch (cmd) {
     case gc_end_match : st = g_end_match;
       //scrollTxtAddTxtLine("g_end_match");
       break;
     case gc_next :
     default : 
       st = g_intro;
       //scrollTxtAddTxtLine("(nextset)");
     }
     break;
   case g_end_match: st = g_splash;
     //scrollTxtAddTxtLine("g_splash");
     break;
   default:
     break;
   }
   ga->step = st;
 }

 /*
  * GLOBALs
  */
 GAME Game;
 GAME_PARAMS GameParams;
 PARTICLE Particle[N_PARTICLE];
 PLAYER Player[N_PLAYER];
 STATION Station[N_STATION];
 CAMERA Camera;
 SPEED_MARKS SpeedMarks; 
 LIGHT Light;
 static	Uint8	gggg[512];



 /*
  * SUBS
  */

 FIXED dist (FIXED x1, FIXED x2){
   FIXED d = x1-x2;
   return ((d>=toFIXED(0.0))?d:x2-x1);
 }

 /* adds attraction of p2 on p1. p1's speed is modified
  *  ->                                               --->
  *  F  = q1q2.(1/((x2-x1)+(y2-y1)+(z2-z1))^(3/2)).p2p1
  *  --->          ->       ->       ->
  *  p2p1 = (x1-x2)i+(y1-y2)j+(z1-z2)k
  * 
  * for each charge type
  */
 void add_attraction(PARTICLE* p1,PARTICLE *p2, FIXED cf){
   FIXED q1,q2,q1q2;
   FIXED dx,dy,dz;
   FIXED N;

   if (!(p1->can_move)) return;

   q1 = q2 = toFIXED(0.0);

   if ((p1->charged[ct_1]) && (p2->charged[ct_1])){
     q1 = p1->e_charge[ct_1]; 
     q2 = p2->e_charge[ct_1];
   }
   if ((p1->charged[ct_2]) && (p2->charged[ct_2])){
     q1 += p1->e_charge[ct_2]; 
     q2 += p2->e_charge[ct_2];
   }
   if ((p1->charged[ct_3]) && (p2->charged[ct_3])){
     q1 += p1->e_charge[ct_3]; 
     q2 += p2->e_charge[ct_3];
   }

   q1q2 = slMulFX(cf,slMulFX(q1,q2)); 

   if (q1q2 == toFIXED(0.0)) return; 

   dx = p1->pos[X]-p2->pos[X];
   dy = p1->pos[Y]-p2->pos[Y];
   dz = p1->pos[Z]-p2->pos[Z];

   N = slMulFX(dx,dx) + slMulFX(dy,dy) + slMulFX(dz,dz);

   N = slMulFX(N,slMulFX(N,N));
   N = slSquartFX(N);

   if (N != toFIXED(0.0)){  
     p1->speed[X] += slDivFX(N,slMulFX(q1q2,dx));
     p1->speed[Y] += slDivFX(N,slMulFX(q1q2,dy));
     p1->speed[Z] += slDivFX(N,slMulFX(q1q2,dz));
   } else {
     //    scrollTxtAddTxtLine("bing             ");
   }
 }

 void compute_attractions(FIXED cf){
   int i,j;
   // charge_factor
   for (i=0;i<N_PARTICLE;i++){
     for (j=0;j<N_PARTICLE;j++){
       if (i!=j){
	 add_attraction(&(Particle[i]),&(Particle[j]),cf);
       }
     }
   }
 }

 int is_collision(PARTICLE* pa, STATION* st){
   int coll = TRUE;
   /*
     FIXED dz = toFIXED(S_SIZE/40.0);
   
   if ((pa->pos[X]>(st->pos[X]+st->size[X]))
       || (pa->pos[X]<(st->pos[X]-st->size[X]))
       || (pa->pos[Y]>(st->pos[Y]+st->size[Y]))
       || (pa->pos[Y]<(st->pos[Y]-st->size[Y]))
       || (pa->pos[Z]>(st->pos[Z]+dz))
       || (pa->pos[Z]<(st->pos[Z]-dz)))    
     coll = FALSE;
   */
   
   FIXED dz = GameParams.spm_pseudo_speed;
   
   if ((pa->pos[X]>(st->pos[X]+st->size[X]))
       || (pa->pos[X]<(st->pos[X]-st->size[X]))
       || (pa->pos[Y]>(st->pos[Y]+st->size[Y]))
       || (pa->pos[Y]<(st->pos[Y]-st->size[Y]))
       || (pa->pos[Z]>st->pos[Z])
       || (pa->pos[Z]<(st->pos[Z]-dz)))    
     coll = FALSE;

   return coll;    
 }
 void compute_collisions(){
   int i,j;
   for (i=0;(i<GameParams.n_players);i++){
     PLAYER* p = &(Player[i]); // foreach player...
     if (p->state == pl_here){
       PARTICLE* pa = p->particle; // ... and player's particle (assumes it exists)     
       for (j=0;(j<N_STATION);j++){ // foreach station
	 STATION* st = &(Station[j]);
	 if (is_collision(pa,st)){
	   if ((st->type == for_all) || ((&(Player[st->num]))== p)){ // station ok for the player <p>
	     player_do_some_good(p);
	     sound_play(snd_good);
	   } else {
	     player_hurt(p);
	     sound_play(snd_bad);
	   }
	   station_leave(st); // next state is "leaving"
	 }
       }
     }
   }
 }

 void set_pseudo_speed(FIXED dz){
   int i;
   // game params
   GameParams.spm_pseudo_speed = dz;
   // speed marks
   speed_marks_set_step_size(&SpeedMarks,dz);
   // stations
   for (i=0;i<N_STATION;i++){
     STATION* st = &(Station[i]);
     station_set_pseudo_speed(st,dz);
   }
 }

 void init_particules(){
   int i=0;

   // human particles 
   particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR1,&PD_LOC_PLAN1,pas_player);
   if (i<N_PLAYER) particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR2,&PD_LOC_PLAN2,pas_player);
   if (i<N_PLAYER) particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR3,&PD_LOC_PLAN3,pas_player);
   if (i<N_PLAYER) particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR4,&PD_LOC_PLAN4,pas_player);
   if (i<N_PLAYER) particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR5,&PD_LOC_PLAN5,pas_player);
   if (i<N_PLAYER) particle_init(&(Particle[i++]),(XPDATA*)&PD_PAR6,&PD_LOC_PLAN6,pas_player);

   // machine
   for (i=N_PLAYER;i<N_PARTICLE;i++){

     PARTICLE * pa;
     pa = &(Particle[i]);
     // particle_init(pa,&PD_PARn,(PDATA*)NULL,pas_attractor);
     particle_init(pa,xpdata_attractor,(PDATA*)NULL,pas_attractor);
     pa->display_tail = FALSE;   
     pa->can_move = FALSE;

     pa->charged[ct_1] = FALSE; 
     particle_set_charge(pa,ct_1,toFIXED(-1.0));
     pa->charged[ct_2] = FALSE; 
     particle_set_charge(pa,ct_2,toFIXED(-1.0));
     pa->charged[ct_3] = FALSE; 
     particle_set_charge(pa,ct_3,toFIXED(-1.0));

     switch (i) {
     case N_PLAYER   : pa->charged[ct_1] = TRUE; break;
     case N_PLAYER+1 : pa->charged[ct_2] = TRUE; break;
     case N_PLAYER+2 :  
     default: 
       pa->charged[ct_3] = TRUE;
     } 
     pa->pos[X] = minmaxrnd(S_Xm,S_XM);
     pa->pos[Y] = minmaxrnd(S_Ym,S_YM);
     pa->pos[Z] = minmaxrnd(S_Zm,S_ZM);    
   }
 }

 void init_stations(){
   int i;
   FIXED sp = GameParams.spm_pseudo_speed;

   // stations for each player
   for (i=0;((i<N_PLAYER) &&  (i<N_STATION));i++){
     switch (i) {
     case 0 : station_init(&Station[i], for_one, i, &PD_STATION1,&PD_STATION1_line,sp);break;
     case 1 : station_init(&Station[i], for_one, i, &PD_STATION2,&PD_STATION2_line,sp);break;
     case 2 : station_init(&Station[i], for_one, i, &PD_STATION3,&PD_STATION3_line,sp);break;
     case 3 : station_init(&Station[i], for_one, i, &PD_STATION4,&PD_STATION4_line,sp);break;
     case 4 : station_init(&Station[i], for_one, i, &PD_STATION5,&PD_STATION5_line,sp);break;
     case 5 : station_init(&Station[i], for_one, i, &PD_STATION6,&PD_STATION6_line,sp);break;
     }   
   }
   // station for all
   i = N_STATION-1;
   station_init(&Station[i], for_all, i, &PD_STATION7,&PD_STATION7_line,sp);

 }

// must assign a pad after init_players 
void init_players(){
   int i;
   for (i=0;i<N_PLAYER;i++){ 
     player_init(&(Player[i]),-1,&(Particle[i])); // ahaha.
   }
 }

 /*
  */

void obey_players_in_start(int* stop){
  int n_pad,j,i;
  int np = GameParams.n_players;
  (*stop) = FALSE;
  
  // press a to join
  for (n_pad=0;n_pad<16;n_pad++) {  
    int padData = Smpc_Peripheral[n_pad].data;
    if (!(PER_DGT_TA & padData)) { // pad <n_pad> presses A
      // add player if : pad <n_pad> is not attributed
      int exists = FALSE;     
      for (j=0;j<np;j++){
	PLAYER* pl = &(Player[j]);
	exists = ((exists) || (n_pad == pl->pad_num));     
      }
      // first available player.
      if (!(exists)){
	PLAYER* pl = &(Player[np]);
	sound_play(snd_birth);
	particle_randomize_pos(pl->particle);
	pl->pad_num = n_pad;
	pl->wants_to_come = TRUE;
	np++;
	//scrollTxtAddTxtLine("ajout");
      }
    }
  }

  // press start to begin.
  for (n_pad=0;n_pad<16;n_pad++) {   
    int padData = Smpc_Peripheral[n_pad].data;
    if ((!(PER_DGT_ST & padData))&& np>1)
      (*stop) = TRUE;
  }

  // nb rounds
  
  GameParams.n_players = np;

  // nb parties

  for (i=0;i<GameParams.n_players;i++){
    PLAYER* pl = (&Player[i]);     // the player
    int padData = Smpc_Peripheral[pl->pad_num].data;
    int upd = pad_GetUpFront(padData,pl->last_pad);     
    if (!(PER_DGT_KL & upd)) {
      GameParams.n_set -= 1;
      sound_play(snd_bing);
    }
    if (!(PER_DGT_KR & upd)) {
      GameParams.n_set += 1;
      sound_play(snd_bing);
    }
    if (GameParams.n_set>666) GameParams.n_set = 11;
    if (GameParams.n_set<1) GameParams.n_set = 0;
    pl->last_pad = padData;
  }
}

void obey_players_in_match(CAMERA* ca){
  int i;

  Uint16 padData, u_padData, d_padData;

  for (i=0;i<GameParams.n_players;i++){
    PLAYER* pl = (&Player[i]);     // the player
    PARTICLE *p = (PARTICLE*)NULL; // his particle

    padData = Smpc_Peripheral[pl->pad_num].data;
    u_padData = pad_GetUpFront(padData,pl->last_pad);     
    d_padData = pad_GetDownFront(padData,pl->last_pad);     

    switch (pl->state){
    case pl_away :
      /*     if (!(PER_DGT_ST & u_padData))
	pl->wants_to_come = TRUE;
      */
      break;
    case pl_here : 
      /*
	if (!(PER_DGT_ST & u_padData)) 
	pl->wants_to_leave = TRUE; 
      */

      // charge
      p = pl->particle;
      particle_set_is_charged(p,ct_1,((!(PER_DGT_TA & padData))||(!(PER_DGT_TX & padData))));
      particle_set_is_charged(p,ct_2,((!(PER_DGT_TB & padData))||(!(PER_DGT_TY & padData))));
      particle_set_is_charged(p,ct_3,((!(PER_DGT_TC & padData))||(!(PER_DGT_TZ & padData))));

      particle_set_charge(p,ct_1,toFIXED(0.0));
      if (!(PER_DGT_TA & padData)) particle_set_charge(p,ct_1,toFIXED(-1.0));
      if (!(PER_DGT_TX & padData)) particle_set_charge(p,ct_1,toFIXED(+1.0));
      particle_set_charge(p,ct_2,toFIXED(0.0));
      if (!(PER_DGT_TB & padData)) particle_set_charge(p,ct_2,toFIXED(-1.0));
      if (!(PER_DGT_TY & padData)) particle_set_charge(p,ct_2,toFIXED(+1.0));
      particle_set_charge(p,ct_3,toFIXED(0.0));
      if (!(PER_DGT_TC & padData)) particle_set_charge(p,ct_3,toFIXED(-1.0));
      if (!(PER_DGT_TZ & padData)) particle_set_charge(p,ct_3,toFIXED(+1.0));

      // display loc ?
      if (!(PER_DGT_KU & d_padData)) particle_switch_display_loc(p);

      // camera source
      if (!(PER_DGT_TL & padData)){
	int idx = tail_get_last(&(p->tail));
	if (idx != -1){
	  camera_setExternalSource(ca,p->tail.pos[idx]);
	  camera_useExternalSource(ca);
	  camera_setMoveTime(ca,toFIXED(10.0));
	}
      }
      // camera destination
      if (!(PER_DGT_TR & padData)){
	camera_setExternalDest(ca,p->pos);
	camera_useExternalDest(ca);
	camera_setMoveTime(ca,toFIXED(20.0));
      }
      break;

      break;
    default: break;
    }

    pl->last_pad = padData;       
  } 
}

void display_all(CAMERA* ca,LIGHT* light){
  int i;
  //static ANGLE ang[] = {0x0000,0x0000,0x0000};
   
  slPushUnitMatrix();
  {
    //l[X]+= slRandom();l[Y]+=slRandom();l[Z]+=slRandom();

    slLight(light->v);
    slLookAt(camera_getSource(ca), camera_getDest(ca), DEGtoANG(0.0));

    slPushMatrix();
    {
      //      slRotX(ang[X]); ang[X] += 0x000E;     
      //      slRotY(ang[Y]); ang[Y] += 0x000f;   
      //      slRotZ(ang[Z]); ang[Z] += 0x000a;   
      //arena_display();
      for (i=0;i<N_PARTICLE;i++) particle_display(&(Particle[i]));
      for (i=0;i<N_STATION;i++)  station_display(&(Station[i]));
      speed_marks_display(&SpeedMarks);
      for (i=0;i<GameParams.n_players;i++) life_bar_display(i, Player[i].energy,GameParams.n_players);
    }   
    slPopMatrix();
  }
  slPopMatrix();     
}		   

/*
 * global parameters are set according to the control
 * pad <pad_num> (the game params controler)
 */
void obey_control_pad(int padnum){
  Uint16 padData;

  padData = Smpc_Peripheral[padnum].data;

  if (!(PER_DGT_TA & padData)) {            // a -> rugosity
    if (!(PER_DGT_KU & padData)) 
      GameParams.rugosity += 0x0000100;
    else if  (!(PER_DGT_KD & padData))
      GameParams.rugosity -= 0x0000100;
  } else if  (!(PER_DGT_TB & padData)) {    // b -> charge_factor
    if (!(PER_DGT_KU & padData)) 
      GameParams.charge_factor += 0x0000100;
    else if  (!(PER_DGT_KD & padData))
      GameParams.charge_factor -= 0x0000100;
  } else if  (!(PER_DGT_TC & padData)){     // c -> gravity
    if (!(PER_DGT_KU & padData))
      GameParams.gravity +=  0x0000100;
    else if  (!(PER_DGT_KD & padData))
      GameParams.gravity -=  0x0000100;
  } else if  (!(PER_DGT_TX & padData)){     // x -> pseudo speed 
    if (!(PER_DGT_KU & padData))
      set_pseudo_speed(GameParams.spm_pseudo_speed + 0x1000);
    else if  (!(PER_DGT_KD & padData))
      set_pseudo_speed(GameParams.spm_pseudo_speed - 0x1000);

  } else if (!(PER_DGT_TY & padData)){     // y -> Size of pseudo speed elems. 
    if (!(PER_DGT_KU & padData))
      GameParams.spm_size_z +=  0x1000;
    else if  (!(PER_DGT_KD & padData))
      GameParams.spm_size_z -=  0x1000;
    speed_marks_set_sizeZ(&SpeedMarks,GameParams.spm_size_z);

  } else if (!(PER_DGT_TZ & padData)){     // y -> Size of pseudo speed elems. 
    int i;
    for (i=N_PLAYER;i<N_PARTICLE;i++){
      particle_randomize_pos(&Particle[i]);
    }
  } else  {
    if (!(PER_DGT_TL & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[Z] += toFIXED(40.0);
    }
    if (!(PER_DGT_TR & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[Z] -= toFIXED(5.0);
    }  
    if (!(PER_DGT_KL & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[X] += toFIXED(5.0);
    }
    if (!(PER_DGT_KR & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[X] -= toFIXED(20.0);
    }  
    if (!(PER_DGT_KU & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[Y] -= toFIXED(20.0);
    }
    if (!(PER_DGT_KD & padData)) {
      FIXED *p;
      p = camera_getInternalSource(&Camera);
      p[Y] += toFIXED(20.0);
    }  
  }
  if (GameParams.rugosity > toFIXED(1.0)) GameParams.rugosity = toFIXED(1.0);
  if (GameParams.rugosity < toFIXED(0.0)) GameParams.rugosity = toFIXED(0.0);
  //  if (GameParams.gravity < toFIXED(0.0)) GameParams.gravity = toFIXED(0.0);
  if (GameParams.charge_factor < toFIXED(0.0)) GameParams.charge_factor = toFIXED(0.0);

  if (!(PER_DGT_ST & padData)) GameParams.display = !(GameParams.display );
}

void dsplyLicense() {
  scrollTxtAddTxtLine("~Le sPin~                 ");
  scrollTxtAddTxtLine("another multiplayer game  ");
  scrollTxtAddTxtLine("rel. under the GNU/GPL    ");
  scrollTxtAddTxtLine("[2003]                    ");
}

void dsplyTitle(int display){
  int i = 8;

  i = 8;
  if (display) {
    slCurColor(i);
    slPrint("                        *         ",slLocate(4,i++));  
    slCurColor(i);
    slPrint(" |         ___ ______  _  __   _  ",slLocate(4,i++));  
    slCurColor(i);
    slPrint(" | e      //_   //_// // //|| //  ",slLocate(4,i++)); 
    slCurColor(i);
    slPrint(" |       __//  //    // // ||//   ",slLocate(4,i++));  
    slCurColor(i);
    slPrint(" --     /__/  //    // //  |//    ",slLocate(4,i++));  
    slCurColor(i);
    i+=5;
    slCurColor(i);
    slPrint("press a start button to begin t   ",slLocate(5,i++));  
    slCurColor(i);
    slPrint("               -o play the game   ",slLocate(5,i++));  
    slCurColor(0);
  } else {
    // ... 
    slPrint("                                  ",slLocate(4,i++));  
    slPrint("                                  ",slLocate(4,i++));  
    slPrint("                                  ",slLocate(4,i++)); 
    slPrint("                                  ",slLocate(4,i++)); 
    slPrint("                                  ",slLocate(4,i++));  
    i+=5;
    slPrint("                                  ",slLocate(5,i++));  
    slPrint("                                  ",slLocate(5,i++));      
  }
}

/*
 * debug
 */

void dbg_display_player_states(){
  int i;
  int ybase = 0;
  player_state st;
  FIXED nrj;
  for (i=0;i<GameParams.n_players;i++){
    slPrint("        ",slLocate(20,ybase+i));
    st = Player[i].state;
    nrj = Player[i].energy;

    /*    switch (st) {
    case pl_coming :slPrint("coming  ",slLocate(10,ybase+i));
      break;
    case pl_here :  slPrint("here    ",slLocate(10,ybase+i));
      break;
    case pl_leaving:slPrint("leaving ",slLocate(10,ybase+i));
      break;
    case pl_dead:   slPrint("dead    ",slLocate(10,ybase+i));
      break;
    case pl_away  : slPrint("away    ",slLocate(10,ybase+i));
      break;  
    default :       slPrint("erreur  ",slLocate(10,ybase+i));
    }
    */
    slPrintFX(nrj,slLocate(18,ybase+i));
  }
}
/*
 *  SUBMAINS
 */

void sub_splash (){
  int more = TRUE;
  dsplyTitle(TRUE);    
  sound_play(snd_lespin);

  while (more){
    Uint16 padData;
    int i;

    //   dbg_display_player_states();
    
    for (i=0;i<16;i++) {
      slRandom(); // init random function for more random
      padData = Smpc_Peripheral[i].data;
      if (!(PER_DGT_ST & padData)) more = FALSE;
    }
    slSynch();
  }

  // fin.
  dsplyTitle(FALSE);
  sound_play(snd_good);

  game_set_step(&Game,gc_next);
}

void sub_start (){
  int i;
  int next = FALSE;

  //dbg_display_player_states();

  // select players
  obey_players_in_start(&next);
  // display

  slPrint("press <A> to JOIN ...",slLocate(10,5));   
  if (GameParams.n_players>1) {
    slPrint("... and <start> to START",slLocate(9,20));
    slPrint("Rounds : ",slLocate(10,1));   
    slPrintFX(toFIXED(1.0*(GameParams.n_set+1)),slLocate(20,1));   
  }

  slPushUnitMatrix();
  {
      FIXED light[] = {minmaxrnd(toFIXED(0.2),toFIXED(2.0)),
		       minmaxrnd(toFIXED(0.2),toFIXED(2.0)),
		       minmaxrnd(toFIXED(0.2),toFIXED(2.0))};
      slLookAt(camera_getSource(&Camera), camera_getDest(&Camera), DEGtoANG(0.0));
      slLight(light);
      for (i=0;i<GameParams.n_players;i++) {
	PARTICLE* pa = Player[i].particle; 
	slPushMatrix();
	{    
	  slTranslate(pa->pos[X],pa->pos[Y],toFIXED(0.0));
	  slPutPolygon((PDATA*)pa->gfx); 
	}
	slPopMatrix();
      }
  }
  slPopMatrix();

  // next
  if (next) {
    slPrint("                                    ",slLocate(10,1));   
    slPrint("                                    ",slLocate(20,1));   
    slPrint("                                    ",slLocate(10,5));   
    slPrint("                                    ",slLocate(9,20));
    game_set_step(&Game,gc_next);
  }
  camera_step(&Camera);
  slSynch();
}

void randomize_game_params_before_set(){
  GAME_PARAMS* gp = &GameParams;
  gp->rugosity = minmaxrnd(toFIXED(0.85),toFIXED(0.96)); 
  gp->charge_factor =  minmaxrnd(toFIXED(0.65),(toFIXED(0.70)));
  gp->gravity = minmaxrnd(toFIXED(-0.05),toFIXED(0.05));
  
  speed_marks_set_sizeZ(&SpeedMarks,minmaxrnd(toFIXED(2.0*S_SIZE),
					      toFIXED((2.0*S_SIZE)/5.0)));
  set_pseudo_speed(minmaxrnd(toFIXED(S_SIZE/100.0),(toFIXED(S_SIZE/10.0))));
  light_randomize(&Light);
}

// for easier moves
void reorder_attractors(FIXED* oa, FIXED* ob, FIXED* oc){
  int i;
  FIXED *oap, *obp, *ocp; 
  FIXED tmpa[XYZ];
  FIXED tmpb[XYZ];
  FIXED tmpc[XYZ];

  // third point
  oc[X]=slMulFX(toFIXED(-1.0),oa[X]+ob[X]);
  oc[Y]=slMulFX(toFIXED(-1.0),oa[Y]+ob[Y]);
  oc[Z]=slMulFX(toFIXED(-1.0),oa[Z]+ob[Z]); 

  // scr limits 
  if (oc[X]>S_XM) oc[X]=S_XM; if (oc[X]<S_Xm) oc[X]=S_Xm;
  if (oc[Y]>S_YM) oc[Y]=S_YM; if (oc[Y]<S_Ym) oc[Y]=S_Ym;
  if (oc[Z]>S_ZM) oc[Z]=S_ZM; if (oc[Z]<S_Zm) oc[Z]=S_Zm;

  // reorder
  oap=oa;obp=ob;ocp=oc; 
  if ((oa[X]<oc[X]) && (oc[X]<ob[X])) {oap=oa;obp=oc;ocp=ob;}
  if ((ob[X]<oa[X]) && (oa[X]<oc[X])) {oap=ob;obp=oa;ocp=oc;}
  if ((ob[X]<oc[X]) && (oc[X]<oa[X])) {oap=ob;obp=oc;ocp=oa;}
  if ((oc[X]<oa[X]) && (oa[X]<ob[X])) {oap=oc;obp=oa;ocp=ob;}
  if ((oc[X]<ob[X]) && (ob[X]<oa[X])) {oap=oc;obp=ob;ocp=oa;}
  
  for (i=0;i<XYZ;i++) tmpa[i] = oap[i];
  for (i=0;i<XYZ;i++) tmpb[i] = obp[i];
  for (i=0;i<XYZ;i++) tmpc[i] = ocp[i];
  
  for (i=0;i<XYZ;i++) oa[i] = tmpa[i];
  for (i=0;i<XYZ;i++) ob[i] = tmpb[i];
  for (i=0;i<XYZ;i++) oc[i] = tmpc[i];
}
	
void sub_intro(){

  int i;
  int more = 500;
  int allhere = FALSE;

  camera_setInternalSource(&Camera,GameParams.camera_dflt_intro_src_pos); 
  camera_setInternalDest(&Camera,GameParams.camera_dflt_intro_dst_pos); 
  camera_setMoveTime(&Camera,toFIXED(1.0));
  camera_step(&Camera); // camera is directly set at             ^
  camera_setInternalSource(&Camera,GameParams.camera_dflt_src_pos); 
  camera_setInternalDest(&Camera,GameParams.camera_dflt_dst_pos);
  camera_setMoveTime(&Camera,toFIXED(500.0));

  randomize_game_params_before_set();

  sound_play(snd_good);
  
  while ((!allhere)||((more--)>0)){

    allhere = TRUE;
    for (i=0;i<GameParams.n_players;i++){
      allhere = (allhere && (Player[i].state == pl_here));
    }

    for (i=0;i<GameParams.n_players;i++) {
      player_step_sm(&(Player[i]));
    }  
  
    display_all(&Camera,&Light);
    camera_step(&Camera);
    scrollTxtDisplay();
    scrollTxtSchedule();    
    
    for (i=N_PLAYER;i<N_PARTICLE;i++){
      particle_randomize_pos(&Particle[i]);
    }
    // ici, on positionne le troisieme machin magntique en fonction des deux premiers
    if ((N_PLAYER+2)<N_PARTICLE){
      reorder_attractors(Particle[N_PLAYER].pos,Particle[N_PLAYER+1].pos,Particle[N_PLAYER+2].pos);
    }
    light_move(&Light);

    //dbg_display_player_states();
    
    if (more & 0x1) slSynch();
  }
  game_set_step(&Game,gc_next);
}

void sub_in_match (){
  int end = FALSE;

  while (!(end)) {
    int i;
    light_move(&Light);

    //dbg_display_player_states();

    for (i=0;i<N_PARTICLE;i++) {
      PARTICLE* pa = &(Particle[i]);
      TAIL* ta = &(pa->tail);
      FIXED wi[] = {toFIXED(0.0),toFIXED(0.0),slMulFX(toFIXED(-1.0),GameParams.spm_pseudo_speed)};
      particle_applyRugosity(pa,GameParams.rugosity);
      particle_applyGravity(pa,GameParams.gravity);
      tail_apply_wind(ta,wi);
    }
      
    // ask reset cam (must be called before obey_player)
    camera_useInternalSource(&Camera);
    camera_useInternalDest(&Camera);   
    camera_oscillate(&Camera);

    obey_players_in_match(&Camera);  
    

    compute_attractions(GameParams.charge_factor);  
    compute_collisions();
    
    for (i=0;i<GameParams.n_players;i++){     
      PLAYER* p = &(Player[i]);
      if (player_is_here(p)) {
	PARTICLE* pa = p->particle;
	particle_move(pa);
      }
    }
    
    display_all(&Camera,&Light);
    
    // prints
    //
    game_params_print(&(GameParams));
    scrollTxtDisplay();
    scrollTxtSchedule();
    
    // debug
    obey_control_pad(15);
    
    // sm !
    for (i=0;i<N_STATION;i++) {
      station_step_sm(&(Station[i]));
    }
    for (i=0;i<GameParams.n_players;i++) {
      player_step_sm(&(Player[i]));
    }     
    camera_step(&Camera);
    speed_marks_step(&SpeedMarks);

    // end ?
    
    end = GameParams.n_players;
    
    for (i=0;i<GameParams.n_players;i++){   
      end -= (Player[i].state == pl_away)?1:0;
    }
      
    end = (end>1)?FALSE:TRUE; // while at least two players remaining...
    
    //    if (end) scrollTxtAddTxtLine("--tous morts");
    
    //    end = (end || (!(PER_DGT_ST & Smpc_Peripheral[0].data)));
    slSynch();
  }
  game_set_step(&Game,gc_next);
}

void display_outro_particles(max_point,winner){
  static ANGLE ang= 0x0;
  static ANGLE acam= 0x0;
  static FIXED sizeWinner = toFIXED(1.0);
  static ANGLE dacam = DEGtoANG(0.2);
  static FIXED dsizeWinner = toFIXED(0.2);

  ANGLE ang2;
  int i;

  ang += DEGtoANG(5.0);

  acam += dacam;
  if (acam>DEGtoANG(50.0)) dacam = DEGtoANG(-0.2);
  if (acam<DEGtoANG(-100.0)) dacam = DEGtoANG(0.1);

  sizeWinner += dsizeWinner;
  if (sizeWinner>toFIXED(3.0)) dsizeWinner = toFIXED(-0.05);
  if (sizeWinner<toFIXED(1.0)) dsizeWinner = toFIXED(+0.1);

  slPushUnitMatrix();
  {
    FIXED light[] = {minmaxrnd(toFIXED(0.2),toFIXED(2.0)),
		     minmaxrnd(toFIXED(0.2),toFIXED(2.0)),
		     minmaxrnd(toFIXED(0.2),toFIXED(2.0))};
    camera_oscillate(&Camera);
    slLookAt(camera_getSource(&Camera), camera_getDest(&Camera), acam);
    slLight(light);
    
    ang2 = 0x0;
    for (i=0;i<GameParams.n_players;i++){
      PLAYER* pl   = &(Player[i]);
      PARTICLE* pa = Player[i].particle; 
      int pp = pl->points;
      FIXED rel_point;      

      ang2 += DEGtoANG(20.0);

      rel_point = toFIXED(1.0)-slDivFX(toFIXED(max_point*1.0),toFIXED(pp*1.0));
      
      slPushMatrix();
      {
	slRotZ(ang+ang2);
	slPushMatrix();
	{
	  slTranslate(slMulFX(rel_point,S_XM),toFIXED(0.0),toFIXED(0.0));
	  if (i == winner){ // winner ?
	    slPushMatrix();
	    {
	      slRotZ(ang-ang2);
	      slPushMatrix();
	      {
		slScale(sizeWinner,sizeWinner,toFIXED(1.0));
		slPutPolygon((PDATA*)pa->gfx); 
	      }
	      slPopMatrix();
	    }
	    slPopMatrix();
	  } else{            // not winner
	    slPutPolygon((PDATA*)pa->gfx); 
	  }
	}
	slPopMatrix();
      }
      slPopMatrix();
    }
  }
  slPopMatrix();
}


/* <dsply> = true -> draws  score screen
 *         = false  -> erases score screen
 */
void display_inter_scores(int dsply,int winner, int maxpoint){
  // display
  if (dsply){

    slPrintFX(toFIXED((GameParams.cur_set+1)*1.0),slLocate(17,4));  
    slPrintFX(toFIXED((GameParams.n_set+1)*1.0),slLocate(25,4));   

    slPrint("Round ",slLocate(13,4));      
    slPrint(" of ",slLocate(23,4));     
    slPrint("      ",slLocate(31,4));    

    slPrint("(press A)",slLocate(30,10));       
    display_outro_particles(maxpoint, winner);     
  } else { 
    // clean !
    slPrint("                            ",slLocate(10,4));
    slPrint("                            ",slLocate(17,4));
    slPrint("                            ",slLocate(23,4));        
    slPrint("                            ",slLocate(27,4));  
    slPrint("                            ",slLocate(30,10));
  }
}

/* <dsply> = true -> draws  score screen
 *         = false  -> erases score screen
 */
void display_final_scores(int dsply, int winner, int maxpoint){

  if (dsply){
    slPrint("Game Over",slLocate(10,4));        // "round over"
    slPrint("(press B)",slLocate(30,25));       
    /*    for (i=0;i<GameParams.n_players;i++){
      slPrintFX(toFIXED(Player[i].points*1.0),slLocate(15,12+i)); // points
    }
    */
    display_outro_particles(maxpoint, winner);     
  } else { 
    // clean !
    slPrint("                            ",slLocate(10,4));
    slPrint("                            ",slLocate(30,25));
    slSynch();
  }
}


void sub_outro (){
  
  int i;
  int more = TRUE;
  int counter = 150;
  int winner = -1;
  int max_point=0; 

  // winner ?
  for (i=0;i<GameParams.n_players;i++){
    winner = (Player[i].state == pl_here)?i:winner;
  }
  // update score
  if (winner>=0) Player[winner].points +=1;

  // max points
  for (i=0;i<GameParams.n_players;i++){
    int pp = Player[i].points;
    max_point = (max_point>pp)?max_point:pp;
  }

  // display score
  while (more){

    display_inter_scores(TRUE,winner,max_point);
    //dbg_display_player_states();

    if (counter < 0){
      Uint16 pad;
      for (i=0;i<16;i++) {
	pad = Smpc_Peripheral[i].data;
	if (!(PER_DGT_TA &  pad)) more = FALSE;
      }
    }else{
      counter--;
    }

    slSynch();
  }
  sound_play(snd_good);
  display_inter_scores(FALSE,0,0);
  
  randomize_game_params_before_set();
   
 
  
  // everybody comes back
  for (i=0;i<GameParams.n_players;i++){
    Player[i].wants_to_come = TRUE;
    player_step_sm(&Player[i]);
    Player[i].energy = toFIXED(10.0);
  }

  if (GameParams.cur_set == GameParams.n_set){
    game_set_step(&Game,gc_end_match);
  } else {
    GameParams.cur_set+=1;
    game_set_step(&Game,gc_next);
  }
  
}

void sub_end_match (){
  int more = TRUE;
  int winner = -1;
  int max_point = GameParams.cur_set; // max POSSIBLE points
  int tmp = 0;
  int j;

  // winner has max points
  for (j=0;j<GameParams.n_players;j++){
    if (Player[j].points>tmp){
      winner = j;
      tmp = Player[j].points;
    }
  }

  while (more){
    Uint16 padData;
    int i;
    
    //dbg_display_player_states();

    for (i=0;i<16;i++) {
      padData = Smpc_Peripheral[i].data;
      if (!(PER_DGT_TB & padData)) more = FALSE;
    }

    display_final_scores(TRUE,winner,max_point);
    //    scrollTxtSchedule();
    //    scrollTxtDisplay();
    slSynch();
  }
  display_final_scores(FALSE,0,0);
  game_set_step(&Game,gc_next);
}

void init_all(){
  game_init(&Game);
  game_params_init(&GameParams);
  camera_init(&Camera,GameParams.camera_dflt_src_pos,GameParams.camera_dflt_dst_pos);
  speed_marks_init(&SpeedMarks, GameParams.spm_size_z, GameParams.spm_pseudo_speed);
  init_particules();
  init_stations();
  init_players(); 
  scrollTxtInit();
  light_init(&Light);
}
/*
 * MAIN
 */
void ss_main() {
  int first = TRUE;
  slInitSystem(TV_320x240,NULL,1); 
  slWindow(0,0,319,223,0xfffe,160,112);
  slZdspLevel(7);  // max should be 3
  slInitGouraud(grtbl_CONCAVE1,160*1, GRtbl(0),gggg);
  slSetAmbient(C_RGB(2,2,2));
  slSetGouraudColor(C_RGB(9,2,5));

  init_sound();
 
  // action
  while (-1) {
    switch (Game.step){
    case g_splash :
      init_all();
      sub_splash();
      if (first) {
	dsplyLicense();
	first = FALSE;
      }
      break;
    case g_start : sub_start();  break;
    case g_intro : sub_intro();  break;
    case g_in_match : sub_in_match();  break;
    case g_outro : sub_outro();  break;
    case g_end_match: sub_end_match(); break;
    }
  }     
}

//---------------------------------------------------------------------------

  /*	TV_320x224 , TV_320x240 ,TV_352x224 ,TV_352x240 ,TV_640x224 ,
	TV_640x240 ,TV_704x224 ,TV_704x240 ,TV_320x448 , TV_320x480 ,
	TV_352x448 ,TV_352x480 ,TV_640x448 ,TV_640x480 ,TV_704x448 ,TV_704x480   */
   

/*  Sound library */
/*
    extern  void   slInitSound(Uint8 * , Uint32 , Uint8 * , Uint32) ;
    extern  Bool   slBGMOn(Uint16 , Uint8 , Uint8 , Uint8) ;
    extern  Bool   slBGMPause() ;
    extern  Bool   slBGMCont() ;
    extern  Bool   slBGMOff() ;
    extern  Bool   slBGMFade(Uint8 , Uint8) ;
    extern  Bool   slBGMTempo(Sint16) ;
#define	    slBGMStat()		slSequenceStat(0)
    extern  Bool   slSoundAllOff() ;
    extern  Bool   slSoundAllPause() ;
    extern  Bool   slSoundAllCont() ;
    extern  Bool   slDSPOff() ;
    extern  Bool   slSndVolume(Uint8) ;
    extern  Uint8  slSequenceOn(Uint16 , Uint8 , Uint8 , Sint8) ;
    extern  Bool   slSequenceOff(Uint8) ;
    extern  Bool   slSequenceFade(Uint8 , Uint8 , Uint8) ;
    extern  Bool   slSequenceTempo(Uint8 , Sint16) ;
    extern  Bool   slSequencePause(Uint8) ;
    extern  Bool   slSequenceCont(Uint8) ;
    extern  Bool   slSequencePan(Uint8 , Sint8) ;
    extern  Bool   slSequenceStat(Uint8) ;
    extern  Bool   slSequenceReset(Uint8) ;
    extern  void  *slSndMapChange(Uint8) ;
    extern  void   slWaitSound(void *) ;
    extern  Bool   slCDDAOn(Uint8 , Uint8 , Sint8 , Sint8) ;
    extern  Bool   slCDDAOff() ;
    extern  Uint8  slPCMOn(PCM * , void * , Uint32) ;
    extern  Bool   slPCMOff(PCM *) ;
    extern  Bool   slPCMParmChange(PCM *) ;
    extern  Bool   slPCMStat(PCM *) ;
    extern  Uint8  slSndPCMNum() ;
    extern  Uint8  slSndSeqNum() ;
    extern  Bool   slSndEffect(Uint8) ;
    extern  Bool   slSndMixChange(Uint8 , Uint8) ;
    extern  Bool   slSndMixParmChange(Uint8 , Uint8 , Sint8) ;
    extern  Sint8  slSoundRequest(const char*,...) ;

    extern  Uint8  sddrv_129i[] ;
    extern  Uint32 sizeof_129i ;
*/
/*------------------------------------------------------------------------*/

   /*  Gouraud library */
   /*
     typedef Uint16 GOURAUDTBL[4];

     extern  void slInitGouraud( GOURAUDTBL *, Uint32, Uint32, Uint8 *);
     extern  void slPutPolygonX( XPDATA *, FIXED *);
     extern  void slGouraudTblCopy(void);
     extern  void slSetGouraudTbl( Uint16 *);
     extern  void slSetGouraudColor( Uint16 );
     extern  void slSetAmbient( Uint16);

     extern  void slSetFlatColor( Uint16);

     extern  void slSetDepthLimit( Uint32, Uint16, Uint16 );
     extern  void slSetDepthTbl( Uint16 *, Uint16, Uint16 );

     extern  void slNearClipFlag( Uint32 );
     extern  void slDispPlaneLevel( Sint32 );
     extern  void slWindowClipLevel( Sint32 );
   */
